import com.fishlib.configuration.Configuration;
import com.fishlib.io.log.LogLevel;
import com.fishlib.io.logger.DelegatingLoggerImpl;
import com.fishlib.io.logger.Logger;
import com.fishlib.logger.ProcessUnsafeLogCreator;
import com.fishlib.net.impl.nio.NIODriver;
import com.fishlib.util.process.ProcessEnvironment;
import com.illumon.iris.db.tables.Table;
import com.illumon.iris.db.tables.TableDefinition;
import com.illumon.iris.db.tables.databases.Database;
import com.illumon.iris.db.tables.remote.RemoteDatabase;
import com.illumon.iris.db.tables.remotequery.ContextAwareRemoteQuery;
import com.illumon.iris.db.tables.remotequery.RemoteProcessingRequest;
import com.illumon.iris.db.tables.remotequery.RemoteQueryClient;
import com.illumon.iris.db.tables.utils.DBTimeUtils;
import com.illumon.iris.db.tables.utils.TableTools;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.io.UncheckedIOException;

/**
 * <p>Simple example for executing {@link QueryUtils#getAvgStockPriceTable(Database, String)}
 * against a remote Deephaven database.</p>
 *
 * <p>This example creates a database connection (RemoteDatabase), and uses its methods ad-hoc,
 * comparably to how these methods are used in an Deephaven console session (db.i, db.t, etc).</p>
 *
 * <p>Because the methods are being called against the RemoteDatabase locally in this client application,
 * each operation requires a round-trip between the client and the worker that is associated with the
 * session. For infrequent or very ad-hoc use, this may be fine; for applications where a query
 * will be executed repeatedly, better performance can be achieved by using a ContextAwareRemoteQuery
 * object to allow the method calls to be made as part of a single execution call from the client
 * to the worker. The CARQExample class example performs the same query as this example does,
 * but does it using a ContextAwareRemoteQuery.</p>
 */
public class RemoteDatabaseExample {
    private static final String PROCESS_NAME = "RemoteDatabaseExample";

    private static void prepareEnvironment() {
        // the Configuration object reads the property file, which has named key-value pairs that control the
        // operation of the system.
        final Configuration configuration = Configuration.getInstance();

        // the log object is used by Deephaven components, and produces output in a PROCESS_NAME.log.<date> with
        // a hard link to PROCESS_NAME.log.current (e.g., DemoIntegration.log.current).  The log files are stored
        // in <workspace>/../logs
        final Logger log = new ProcessUnsafeLogCreator(PROCESS_NAME, configuration).createLogger(PROCESS_NAME, LogLevel.INFO);

        // All log output should be sent to the Logger we created on the line above
        DelegatingLoggerImpl.setDelegate(log);

        // The Deephaven system uses a wrapper around Java NIO for most network communication and job scheduling, this
        // initializes the communications subsystem with our log
        NIODriver.init(log);

        // The ProcessEnvironment provides hooks for consistent startup and shutdown initialization across processes.
        ProcessEnvironment.basicInteractiveProcessInitialization(configuration, PROCESS_NAME, log);
    }

    /**
     * Run a sample query to get the average prices of stocks for a given day.
     *
     * @param args 1: The date to use. If not specified, the query will use currentDateNy.
     */
    public static void main(@NotNull final String... args) {
        // Parse parameter values from the command line
        final String date;
        if (args.length > 0) {
            date = args[0];
        } else {
            date = DBTimeUtils.currentDateNy();
        }

        final long overallStartNanos = System.nanoTime();

        // Set up logging and process management
        prepareEnvironment();

        // Create a Deephaven query client
        final RemoteQueryClient rqc;
        try {
            rqc = new RemoteQueryClient();
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to create RemoteQueryClient", e);
        }

        /*
         * Push the QueryUtils class to the worker. This class will be needed by
         * by the worker when executeQuery() is called.
         */
        try {
            rqc.pushClasses(QueryUtils.class);
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to push QueryUtils class", e);
        }

        // Create a connection to the remote query server specified by RemoteQueryDispatcherParameters.host
        final RemoteDatabase db;
        try {
            db = rqc.getRemoteDB(RemoteProcessingRequest.DEFAULT_MAX_HEAP_SIZE_MB, PROCESS_NAME, RemoteProcessingRequest.DEFAULT_WORKER_DEBUG_ENABLED);
        } catch (IOException e) {
            throw new UncheckedIOException("Failed to create RemoteDatabase", e);
        }

        final Table AvgStockPrices = QueryUtils.getAvgStockPriceTable(db, date);

        // Output the results
        TableTools.show(AvgStockPrices);
    }
}